@@ -25,28 +25,19 @@ def divisible_by(val, divisor):
 def cast_tuple(val, num = 1):
     return val if isinstance(val, tuple) else ((val,) * num)
 
-# helper functions
+# deepnet init
 
-class PreNorm(nn.Module):
-    def __init__(self, dim, fn):
-        super().__init__()
-        self.fn = fn
-        self.norm = RMSNorm(dim)
+def deepnorm_init(transformer, beta, module_name_match_list = ['.ff.', '.to_v', '.to_out']):
+    for name, module in transformer.named_modules():
+        if type(module) != nn.Linear:
+            continue
 
-    def forward(self, x, *args, **kwargs):
-        return self.fn(self.norm(x), *args, **kwargs) + x
+        needs_beta_gain = any(map(lambda substr: substr in name, module_name_match_list))
+        gain = beta if needs_beta_gain else 1
+        nn.init.xavier_normal_(module.weight.data, gain = gain)
 
-class PostNorm(nn.Module):
-    def __init__(self, dim, fn, scale_residual = 1):
-        super().__init__()
-        self.fn = fn
-        self.scale_residual = scale_residual
-        self.norm = RMSNorm(dim)
-
-    def forward(self, x, *args, **kwargs):
-        residual = x * self.scale_residual
-        out = self.fn(x, *args, **kwargs) + residual
-        return self.norm(out)
+        if exists(module.bias):
+            nn.init.constant_(module.bias.data, 0)
 
 # normalization
 
@@ -66,6 +57,29 @@ class RMSNorm(nn.Module):
         norm = x.norm(keepdim = True, dim = -1) * self.scale
         return (x / norm.clamp(min = self.eps)) * self.gamma
 
+# pre and post norm residual wrapper modules
+
+class PreNorm(nn.Module):
+    def __init__(self, dim, fn, norm_klass = RMSNorm):
+        super().__init__()
+        self.fn = fn
+        self.norm = norm_klass(dim)
+
+    def forward(self, x, *args, **kwargs):
+        return self.fn(self.norm(x), *args, **kwargs) + x
+
+class PostNorm(nn.Module):
+    def __init__(self, dim, fn, scale_residual = 1, norm_klass = RMSNorm):
+        super().__init__()
+        self.fn = fn
+        self.scale_residual = scale_residual
+        self.norm = norm_klass(dim)
+
+    def forward(self, x, *args, **kwargs):
+        residual = x * self.scale_residual
+        out = self.fn(x, *args, **kwargs) + residual
+        return self.norm(out)
+
 # positional embedding
 
 class RotaryEmbedding(nn.Module):
@@ -93,15 +107,20 @@ def apply_rotary_pos_emb(t, freqs):
 
 # feedforward
 
-def FeedForward(dim, mult = 4, dropout = 0.):
-    inner_dim = int(mult * dim)
+class FeedForward(nn.Module):
+    def __init__(self, dim, mult = 4, dropout = 0.):
+        super().__init__()
+        inner_dim = int(mult * dim)
 
-    return nn.Sequential(
-        nn.Linear(dim, inner_dim),
-        nn.GELU(),
-        nn.Dropout(dropout),
-        nn.Linear(inner_dim, dim)
-    )
+        self.ff = nn.Sequential(
+            nn.Linear(dim, inner_dim),
+            nn.GELU(),
+            nn.Dropout(dropout),
+            nn.Linear(inner_dim, dim)
+        )
+
+    def forward(self, x):
+        return self.ff(x)
 
 # attention
 
@@ -125,20 +144,21 @@ class Attention(nn.Module):
         self.dropout = nn.Dropout(dropout)
 
         self.to_q = nn.Linear(dim, inner_dim, bias = False)
-        self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)
+        self.to_k = nn.Linear(dim, inner_dim, bias = False)
+        self.to_v = nn.Linear(dim, inner_dim, bias = False)
         self.to_out = nn.Linear(inner_dim, dim)
 
         # allowing for attending to nothing (null function)
         # and to save attention from breaking if all retrieved chunks are padded out
-        self.null_kv = nn.Parameter(torch.randn(2, inner_dim)) if null_kv else None
+        self.null_k = nn.Parameter(torch.randn(inner_dim)) if null_kv else None
+        self.null_v = nn.Parameter(torch.randn(inner_dim)) if null_kv else None
 
     def forward(self, x, mask = None, context = None, pos_emb = None):
         b, device, h, scale = x.shape[0], x.device, self.heads, self.scale
 
         kv_input = default(context, x)
 
-        q = self.to_q(x)
-        k, v = self.to_kv(kv_input).chunk(2, dim = -1)
+        q, k, v = self.to_q(x), self.to_k(kv_input), self.to_v(kv_input)
 
         # split heads
 
@@ -158,8 +178,8 @@ class Attention(nn.Module):
 
         # add null key / values
 
-        if exists(self.null_kv):
-            nk, nv = self.null_kv.unbind(dim = 0)
+        if exists(self.null_k):
+            nk, nv = self.null_k, self.null_v
             nk, nv = map(lambda t: repeat(t, '(h d) -> b h 1 d', b = b, h = h), (nk, nv))
             k = torch.cat((nk, k), dim = -2)
             v = torch.cat((nv, v), dim = -2)
@@ -173,7 +193,7 @@ class Attention(nn.Module):
         mask_value = -torch.finfo(sim.dtype).max
 
         if exists(mask):
-            if exists(self.null_kv):
+            if exists(self.null_k):
                 mask = F.pad(mask, (1, 0), value = True)
 
             mask = rearrange(mask, 'b j -> b 1 1 j')
@@ -284,6 +304,7 @@ class Encoder(nn.Module):
         final_norm = True,
         cross_attn_layers = None,
         post_norm = False,
+        norm_klass = RMSNorm,
         scale_residual = 1.
     ):
         super().__init__()
@@ -292,7 +313,7 @@ class Encoder(nn.Module):
         rotary_emb_dim = max(dim_head // 2, MIN_DIM_HEAD)
         self.rotary_pos_emb = RotaryEmbedding(rotary_emb_dim)
 
-        wrapper = partial(PreNorm, dim) if not post_norm else partial(PostNorm, dim, scale_residual = scale_residual)
+        wrapper = partial(PreNorm, dim, norm_klass = norm_klass) if not post_norm else partial(PostNorm, dim, scale_residual = scale_residual, norm_klass = norm_klass)
 
         for layer_num in range(1, depth + 1):
             has_cross_attn = not exists(cross_attn_layers) or layer_num in cross_attn_layers
@@ -303,7 +324,7 @@ class Encoder(nn.Module):
                 wrapper(FeedForward(dim = dim, mult = ff_mult, dropout = ff_dropout)),
             ]))
 
-        self.norm_out = RMSNorm(dim) if final_norm and not post_norm else nn.Identity()
+        self.norm_out = norm_klass(dim) if final_norm and not post_norm else nn.Identity()
 
     def forward(self, x, *, mask = None, chunked_seq):
         device, chunk_size, seq_len = x.device, x.shape[-2], chunked_seq.shape[-2]
@@ -336,6 +357,7 @@ class Decoder(nn.Module):
         cross_attn_layers = None,
         chunk_size = 64,
         post_norm = False,
+        norm_klass = RMSNorm,
         scale_residual = 1.
     ):
         super().__init__()
@@ -344,7 +366,7 @@ class Decoder(nn.Module):
         rotary_emb_dim = max(dim_head // 2, MIN_DIM_HEAD)
         self.rotary_pos_emb = RotaryEmbedding(rotary_emb_dim)
 
-        wrapper = partial(PreNorm, dim) if not post_norm else partial(PostNorm, dim, scale_residual = scale_residual)
+        wrapper = partial(PreNorm, dim, norm_klass = norm_klass) if not post_norm else partial(PostNorm, dim, scale_residual = scale_residual, norm_klass = norm_klass)
 
         self.chunk_size = chunk_size
 
@@ -410,7 +432,8 @@ class RETRO(nn.Module):
         pad_id = 0,
         enc_scale_residual = None,
         dec_scale_residual = None,
-        post_norm = False
+        norm_klass = None,
+        use_deepnet = False
     ):
         super().__init__()
         assert dim_head >= MIN_DIM_HEAD, f'dimension per head must be greater than {MIN_DIM_HEAD}'
@@ -428,9 +451,10 @@ class RETRO(nn.Module):
         # for deepnet, residual scales
         # follow equation in Figure 2. in https://arxiv.org/abs/2203.00555
 
-        if post_norm:
-            enc_scale_residual = default(enc_scale_residual, 0.81 * ((enc_depth ** 4) * dec_depth) ** (-1 / 16))
+        if use_deepnet:
+            enc_scale_residual = default(enc_scale_residual, 0.81 * ((enc_depth ** 4) * dec_depth) ** .0625)
             dec_scale_residual = default(dec_scale_residual, (3 * dec_depth) ** 0.25)
+            norm_klass = nn.LayerNorm
 
         # define encoder and decoders
 
@@ -440,7 +464,8 @@ class RETRO(nn.Module):
             attn_dropout = enc_attn_dropout,
             ff_dropout = enc_ff_dropout,
             cross_attn_layers = enc_cross_attn_layers,
-            post_norm = post_norm,
+            post_norm = use_deepnet,
+            norm_klass = norm_klass,
             scale_residual = enc_scale_residual
         )
 
@@ -451,12 +476,19 @@ class RETRO(nn.Module):
             ff_dropout = dec_ff_dropout,
             cross_attn_layers = dec_cross_attn_layers,
             chunk_size = chunk_size,
-            post_norm = post_norm,
+            post_norm = use_deepnet,
+            norm_klass = norm_klass,
             scale_residual = dec_scale_residual
         )
 
         self.to_logits = nn.Linear(dec_dim, num_tokens)
 
+        # deepnet has special init of weight matrices
+
+        if use_deepnet:
+            deepnorm_init(self.encoder, 0.87 * ((enc_depth ** 4) * dec_depth) ** -0.0625)
+            deepnorm_init(self.decoder, (12 * dec_depth) ** -0.25)
+
     def forward_without_retrieval(
         self,
         seq
